#include "defines.h"

#define REG              0.0001

/*
 * gnsCholeskyDec:
 *
 * Parameters:
 *     inpMatr          I: 6x6 symmetric input matrix
 *     outMatr          O: 6x6 upper triangular output matrix
 *
 * Function:
 *     Performs Cholesky decomposition of inpMatr, i.e. calculates
 *     upper triangular outMatr such that outMatr'*outMatr = inpMatr
 * Returns:
 *     1 if inpMatr is positive definite, i.e. factorization has been done
 *     0 if inpMatr is not positive definite, i.e. factorization has failed
 *
 */

int gnsCholeskyDec(double **inpMatr, double outMatr[MAX_SQR_FILT_LENGTH][MAX_SQR_FILT_LENGTH], int noEq)
{ 
  int 
    i, j, k;     /* Looping Variables */
  double 
    scale;       /* scaling factor for each row */
  double 
    invDiag[MAX_SQR_FILT_LENGTH];  /* Vector of the inverse of diagonal entries of outMatr */


  /*
   *  Cholesky decomposition starts
   */

  for(i = 0; i < noEq; i++)
    for(j = i; j < noEq; j++)
    {
      /* Compute the scaling factor */
      scale=inpMatr[i][j];
      if ( i > 0) for( k = i - 1 ; k >= 0 ; k--)
        scale -= outMatr[k][j] * outMatr[k][i];

      /* Compute i'th row of outMatr */
      if(i==j){
        if(scale <= 0 ){  /* If inpMatr is singular */
          return(0);
        }
        else              /* Normal operation */
          invDiag[i] =  1.0/(outMatr[i][i]=sqrt(scale));
      }
      else{
        outMatr[i][j] = scale*invDiag[i]; /* Upper triangular part          */
        outMatr[j][i] = 0.0;              /* Lower triangular part set to 0 */
      }                    
    }
  
  return(1); /* Signal that Cholesky factorization is successfully performed */
}

/*
 * gnsTransposeBacksubstitution:
 *
 * Parameters:
 *     U          I: a 6x6 upper triangular matrix
 *     rhs        I: a vector of size 6
 *     x          O: vector that satisfies U'*x = rhs
 *
 * Function:
 *     Solves U'*x = rhs for x
 *
 * Returns:
 *     Nothing
 *
 * History
 *     8.4.97         Initial version.
 */

void gnsTransposeBacksubstitution(double U[MAX_SQR_FILT_LENGTH][MAX_SQR_FILT_LENGTH], double rhs[], 
                                  double x[], int order)
{
  int 
    i,j;              /* Looping variables */
  double 
    sum;              /* Holds backsubstitution from already handled rows */

  /* Backsubstitution starts */
  x[0] = rhs[0]/U[0][0];               /* First row of U'                   */
  for (i = 1; i < order; i++){         /* For the rows 1..order-1           */

    for (j = 0, sum = 0.0; j < i; j++) /* Backsubst already solved unknowns */
      sum += x[j]*U[j][i];

    x[i]=(rhs[i] - sum)/U[i][i];       /* i'th component of solution vect.  */
  }
}


/*
 * gnsTransposeBacksubstitution:
 *
 * Parameters:
 *     R          I: a 6x6 upper triangular matrix
 *     z          I: a vector of size 6
 *     A          O: vector that satisfies R*A = z
 *
 * Function:
 *     Solves by backsubstitution the system given by the matrix
 *     equation: R * A = z, with R being upper-triangular.
 *
 * Returns:
 *     Nothing
 *
 */

void gnsBacksubstitution(double R[MAX_SQR_FILT_LENGTH][MAX_SQR_FILT_LENGTH], double z[MAX_SQR_FILT_LENGTH], int R_size, double A[MAX_SQR_FILT_LENGTH])
{
  int
    i, j;

  double
    sum;

  R_size--;

  A[R_size] = z[R_size] / R[R_size][R_size];

  for (i = R_size-1; i >= 0; i--) {

    for (j = i+1, sum = 0.0; j <= R_size; j++)
      sum += R[i][j] * A[j];

    A[i] = (z[i] - sum) / R[i][i];

  }
}

/*
 * gnsSolveByChol:
 *
 * Parameters:
 *     LHS              I: Left Hand Side Matrix(6x6) of least squares eqn.
 *     rhs              I: Right Hand Side vector(6x1) of least squares eqn.
 *     x                O: vector that satisfies LHS*x = rhs
 *
 * Function:
 *     Solves LHS*x = rhs for x. Uses Cholesky decomposition + two
 *     backsubstitutions for the solution. In case LHS is singular, 
 *     regularizes LHS by adding a small positive value to the
 *     diagonals of LHS
 *
 * Returns:
 *     Nothing
 *
 */

int gnsSolveByChol(double **LHS, double *rhs, double *x, int noEq)
{ 
  double
    aux[MAX_SQR_FILT_LENGTH],     /* Auxiliary vector */
    U[MAX_SQR_FILT_LENGTH][MAX_SQR_FILT_LENGTH];    /* Upper triangular Cholesky factor of LHS */
  int
    i, singular;          /* Looping variable */
  
  /* The equation to be solved is LHSx = rhs */
  
  /* Compute upper triangular U such that U'*U = LHS */
  if(gnsCholeskyDec(LHS, U, noEq)){ /* If Cholesky decomposition has been successful */

    singular=1;
    /* Now, the equation is  U'*U*x = rhs, where U is upper triangular
     * Solve U'*aux = rhs for aux
     */
    gnsTransposeBacksubstitution(U, rhs, aux, noEq);         

    /* The equation is now U*x = aux, solve it for x (new motion coefficients) */
    gnsBacksubstitution(U, aux, noEq, x);   

  }
  else{ /* LHS was singular */ 

    singular=0;
    /* Regularize LHS */
    for(i=0; i<noEq; i++)
      LHS[i][i] += REG;

    /* Compute upper triangular U such that U'*U = regularized LHS */
    gnsCholeskyDec(LHS, U, noEq);

    /* Solve  U'*aux = rhs for aux */  
    gnsTransposeBacksubstitution(U, rhs, aux, noEq);   

    /* Solve U*x = aux for x */
    gnsBacksubstitution(U, aux, noEq, x);
  }  
  return(singular);
}


//////////////////////////////////////////////////////////////////////////////////////////


void add_A(double **Amerged, double ***A, int start, int stop, int size)
{ 
  int
    i, j, ind;          /* Looping variable */

  for (i=0; i<size; i++){
    for (j=0; j<size; j++){
      Amerged[i][j]=0;
      for (ind=start; ind<=stop; ind++){
        Amerged[i][j]+=A[ind][i][j];
      }
    }
  }
}

void add_b(double *bmerged, double **b, int start, int stop, int size)
{ 
  int
    i, ind;          /* Looping variable */

  for (i=0; i<size; i++){
    bmerged[i]=0;
    for (ind=start; ind<=stop; ind++){
      bmerged[i]+=b[ind][i];
    }
  }
}
  
double add_pixAcc(double *pixAcc, int start, int stop)
{ 
  int ind;
  double pixAccMerged=0;

  for (ind=start; ind<=stop; ind++){
    pixAccMerged+=pixAcc[ind];
  }
  return(pixAccMerged);
}

double calculateErrorCoeffProvided(double **A, double *b, double *c, int size)
{
  int i, j;
  double error, sum;

  error=0;
  for (i=0; i<size; i++){
    sum=0;
    for (j=0; j<size; j++){
      sum+=A[i][j]*c[j];
    }
    error+=c[i]*sum;
  }

  sum=0;
  for (i=0; i<size; i++){
    sum+=c[i]*b[i];
  }
  error-=(2*sum);

  return(error);
}


double calculateErrorAbs(double **A, double *b, double y, int size)
{
  int i;
  double error, sum;
  static double *c;
  static int first=0;

  if(first == 0)
  {
    first = 1;
    c = (double *) calloc(MAX_SQR_FILT_LENGTH, sizeof(double));
  }

  gnsSolveByChol(A, b, c, size);

  sum=0;
  for (i=0; i<size; i++){
    sum+=c[i]*b[i];
  }
  error=y-sum;

  return(error);
}

double mergeFiltersExoustive(double **yGlobalSeq, double ***EGlobalSeq, double *pixAccGlobalSeq, int intervalBest[NO_VAR_BINS][2], 
                           int sqrFiltLength, int noIntervals)
{
  int first, ind[NO_VAR_BINS], interval[NO_VAR_BINS][2], i;
  double error, errorMin;
  static double *ymerged, **Emerged, pixAcc_merged;
  static int init=0;

  if(init == 0)
  {
    initMatrix_double(&Emerged, MAX_SQR_FILT_LENGTH, MAX_SQR_FILT_LENGTH);
    ymerged = (double *) calloc(MAX_SQR_FILT_LENGTH, sizeof(double));
    init = 1;
  }

  errorMin=0; first=1;

  // Try merging different matrices

  if (noIntervals==2){

    for (ind[0]=1; ind[0]<NO_VAR_BINS; ind[0]++){
      interval[0][0]=0;     interval[0][1]=ind[0]-1;
      interval[1][0]=ind[0];interval[1][1]=NO_VAR_BINS-1;

      error=0;
      for (i=0; i<2; i++){
        add_A(Emerged, EGlobalSeq, interval[i][0], interval[i][1], sqrFiltLength);
        add_b(ymerged, yGlobalSeq, interval[i][0], interval[i][1], sqrFiltLength);
        pixAcc_merged=add_pixAcc(pixAccGlobalSeq, interval[i][0], interval[i][1]);

        error+=calculateErrorAbs(Emerged, ymerged, pixAcc_merged, sqrFiltLength);
      }
      if (error<errorMin || first==1){
        errorMin=error;
        for (i=0; i<noIntervals; i++){
          intervalBest[i][0]=interval[i][0]; intervalBest[i][1]=interval[i][1];
        }
        first=0;
      }
    }
  }
  else if (noIntervals==4){

    for (ind[0]=1; ind[0]<NO_VAR_BINS-2; ind[0]++){
      for (ind[1]=ind[0]+1; ind[1]<NO_VAR_BINS-1; ind[1]++){
        for (ind[2]=ind[1]+1; ind[2]<NO_VAR_BINS; ind[2]++){

          interval[0][0]=0;          interval[0][1]=ind[0]-1;
          for (i=1; i<3; i++){
            interval[i][0]=ind[i-1]; interval[i][1]=ind[i]-1;
          }
          interval[3][0]=ind[2];     interval[3][1]=NO_VAR_BINS-1;

          error=0;
          for (i=0; i<4; i++){
            add_A(Emerged, EGlobalSeq, interval[i][0], interval[i][1], sqrFiltLength);
            add_b(ymerged, yGlobalSeq, interval[i][0], interval[i][1], sqrFiltLength);
            pixAcc_merged=add_pixAcc(pixAccGlobalSeq, interval[i][0], interval[i][1]);

            error+=calculateErrorAbs(Emerged, ymerged, pixAcc_merged, sqrFiltLength);
          }
          if (error<errorMin || first==1){
            errorMin=error;
            for (i=0; i<noIntervals; i++){
              intervalBest[i][0]=interval[i][0]; intervalBest[i][1]=interval[i][1];
            }
            first=0;
          }
        }
      }
    }
  }
  else{

    for (ind[0]=1; ind[0]<NO_VAR_BINS-6; ind[0]++){
      for (ind[1]=ind[0]+1; ind[1]<NO_VAR_BINS-5; ind[1]++){
        for (ind[2]=ind[1]+1; ind[2]<NO_VAR_BINS-4; ind[2]++){
          for (ind[3]=ind[2]+1; ind[3]<NO_VAR_BINS-3; ind[3]++){
            for (ind[4]=ind[3]+1; ind[4]<NO_VAR_BINS-2; ind[4]++){
              for (ind[5]=ind[4]+1; ind[5]<NO_VAR_BINS-1; ind[5]++){
                for (ind[6]=ind[5]+1; ind[6]<NO_VAR_BINS; ind[6]++){

                  interval[0][0]=0;          interval[0][1]=ind[0]-1;
                  for (i=1; i<7; i++){
                    interval[i][0]=ind[i-1]; interval[i][1]=ind[i]-1;
                  }
                  interval[7][0]=ind[6];     interval[7][1]=NO_VAR_BINS-1;

                  error=0;
                  for (i=0; i<8; i++){
                    add_A(Emerged, EGlobalSeq, interval[i][0], interval[i][1], sqrFiltLength);
                    add_b(ymerged, yGlobalSeq, interval[i][0], interval[i][1], sqrFiltLength);
                    pixAcc_merged=add_pixAcc(pixAccGlobalSeq, interval[i][0], interval[i][1]);

                    error+=calculateErrorAbs(Emerged, ymerged, pixAcc_merged, sqrFiltLength);
                  }

                  if (error<errorMin || first==1){
                    errorMin=error;
                    for (i=0; i<noIntervals; i++){
                      intervalBest[i][0]=interval[i][0]; intervalBest[i][1]=interval[i][1];
                    }
                    first=0;
                  }

                }
              }
            }
          }
        }
      }
    }

  }
  return(errorMin);
}




double mergeFiltersGreedy(double **yGlobalSeq, double ***EGlobalSeq, double *pixAccGlobalSeq, int intervalBest[NO_VAR_BINS][2], 
                          int sqrFiltLength, int noIntervals)
{
  int first, ind, ind1, ind2, indexList[NO_VAR_BINS], available[NO_VAR_BINS], i, j, 
    bestToMerge, noRemaining;
  double error, error1, error2, errorMin;
  static double **y_merged, ***E_merged, *pixAcc_merged, *y_temp, **E_temp, pixAcc_temp;
  static int init=0;

  if(init == 0)
  {
    initMatrix_double(&E_temp, MAX_SQR_FILT_LENGTH, MAX_SQR_FILT_LENGTH);
    y_temp = (double *) calloc(MAX_SQR_FILT_LENGTH, sizeof(double));

    initMatrix3D_double(&E_merged, NO_VAR_BINS, MAX_SQR_FILT_LENGTH, MAX_SQR_FILT_LENGTH);
    initMatrix_double(&y_merged, NO_VAR_BINS, MAX_SQR_FILT_LENGTH); 
    pixAcc_merged = (double *) calloc(NO_VAR_BINS, sizeof(double));

    init = 1;
  }

  noRemaining=NO_VAR_BINS;

  for (ind=0; ind<NO_VAR_BINS; ind++){
    indexList[ind]=ind; available[ind]=1;

    pixAcc_merged[ind]=pixAccGlobalSeq[ind];
    for (i=0; i<sqrFiltLength; i++){
      y_merged[ind][i]=yGlobalSeq[ind][i];
      for (j=0; j<sqrFiltLength; j++){
        E_merged[ind][i][j]=EGlobalSeq[ind][i][j];
      }
    }
  }

  // Try merging different matrices
  if (noIntervals==1){

    ind1=0;
    for (ind=1; ind<NO_VAR_BINS; ind++){
      pixAcc_merged[ind1]+=pixAcc_merged[ind];
      for (i=0; i<sqrFiltLength; i++){
        y_merged[ind1][i]+=y_merged[ind][i];
        for (j=0; j<sqrFiltLength; j++){
          E_merged[ind1][i][j]+=E_merged[ind][i][j];
        }
      }
    }

  }
  else{

    while (noRemaining>noIntervals){

      errorMin=0; first=1;
      bestToMerge = 0;
      for (ind=0; ind<noRemaining-1; ind++){

        ind1=indexList[ind];
        ind2=indexList[ind+1];

        error1=calculateErrorAbs(E_merged[ind1], y_merged[ind1], pixAcc_merged[ind1], sqrFiltLength);
        error2=calculateErrorAbs(E_merged[ind2], y_merged[ind2], pixAcc_merged[ind2], sqrFiltLength);

        pixAcc_temp=pixAcc_merged[ind1]+pixAcc_merged[ind2];
        for (i=0; i<sqrFiltLength; i++){
          y_temp[i]=y_merged[ind1][i]+y_merged[ind2][i];
          for (j=0; j<sqrFiltLength; j++){
            E_temp[i][j]=E_merged[ind1][i][j]+E_merged[ind2][i][j];
          }
        }
        error=calculateErrorAbs(E_temp, y_temp, pixAcc_temp, sqrFiltLength)-error1-error2;

        if (error<errorMin || first==1){
          errorMin=error;
          bestToMerge=ind;
          first=0;
        }
      }

      ind1=indexList[bestToMerge];
      ind2=indexList[bestToMerge+1];

      pixAcc_merged[ind1]+=pixAcc_merged[ind2];
      for (i=0; i<sqrFiltLength; i++){
        y_merged[ind1][i]+=y_merged[ind2][i];
        for (j=0; j<sqrFiltLength; j++){
          E_merged[ind1][i][j]+=E_merged[ind2][i][j];
        }
      }

      available[ind2]=0;

      ind=0;
      for (i=0; i<NO_VAR_BINS; i++){
        if (available[i]==1){
          indexList[ind]=i;
          ind++;
        }
      }

      noRemaining--;
    }

  }
  errorMin=0;
  for (ind=0; ind<noIntervals; ind++){
    ind1=indexList[ind];
    errorMin+=calculateErrorAbs(E_merged[ind1], y_merged[ind1], pixAcc_merged[ind1], sqrFiltLength);
  }

  for (ind=0; ind<noIntervals-1; ind++){
    intervalBest[ind][0]=indexList[ind]; intervalBest[ind][1]=indexList[ind+1]-1;
  }
  intervalBest[noIntervals-1][0]=indexList[noIntervals-1]; intervalBest[noIntervals-1][1]=NO_VAR_BINS-1;

  return(errorMin);
}



double findFilterGroupingError(double ***EGlobalSeq, double **yGlobalSeq, double *pixAccGlobalSeq, 
                               int intervalBest[NO_VAR_BINS][2], int sqrFiltLength, int filters_per_fr)
{
  double error;

  // find best filters for each frame group

  error = 0;

  if((filters_per_fr == 2) || (filters_per_fr == 4) || ((filters_per_fr == 8) && (NO_VAR_BINS <= 16)))
    error += mergeFiltersExoustive(yGlobalSeq, EGlobalSeq,  pixAccGlobalSeq, intervalBest, sqrFiltLength, filters_per_fr);
  else
    error += mergeFiltersGreedy(yGlobalSeq, EGlobalSeq, pixAccGlobalSeq, intervalBest, sqrFiltLength, filters_per_fr);

  return(error);

}


void roundFiltCoeff(int *FilterCoeffQuan, double *FilterCoeff, int sqrFiltLength, int factor)
{
  int i;
  double diff; 
  int diffInt, sign; 

  for(i = 0; i < sqrFiltLength; i++)
  {
    sign               = (FilterCoeff[i]>0) ?  1: -1; 
    diff               = FilterCoeff[i]*sign; 
    diffInt            = (int)(diff*(double)factor+0.5); 
    FilterCoeffQuan[i] = diffInt*sign;  
  }
}


double QuantizeIntegerFilterPP(double *filterCoeff, int *filterCoeffQuant, double **E, double *y, int sqrFiltLength, int *weights, int bit_depth)
{
  int factor = (1<<(bit_depth-1)), i, k, minInd, targetCoeffSumInt, diff, quantCoeffSum;
  double targetCoeffSum, errMin, error;
  static int init=0, *filterCoeffQuantMod;

  if(init == 0)
  {
    filterCoeffQuantMod = (int *) calloc(MAX_SQR_FILT_LENGTH, sizeof(int));
    init = 1;
  }

  gnsSolveByChol(E, y, filterCoeff, sqrFiltLength);
  roundFiltCoeff(filterCoeffQuant, filterCoeff, sqrFiltLength, factor);

  targetCoeffSum=0;
  for (i=0; i<sqrFiltLength; i++){
    targetCoeffSum+=(weights[i]*filterCoeff[i]*factor);
  }
  targetCoeffSumInt=ROUND(targetCoeffSum);

  quantCoeffSum=0;
  for (i=0; i<sqrFiltLength; i++){
    quantCoeffSum+=weights[i]*filterCoeffQuant[i];
  }

  while(quantCoeffSum!=targetCoeffSumInt){

    if (quantCoeffSum>targetCoeffSumInt){

      diff=quantCoeffSum-targetCoeffSumInt;
      errMin=0; minInd=-1;
      for (k=0; k<sqrFiltLength; k++){
        if (weights[k]<=diff){
          for (i=0; i<sqrFiltLength; i++){
            filterCoeffQuantMod[i]=filterCoeffQuant[i];
          }
          filterCoeffQuantMod[k]--;
          for (i=0; i<sqrFiltLength; i++){
            filterCoeff[i]=(double)filterCoeffQuantMod[i]/(double)factor;
          }
          error=calculateErrorCoeffProvided(E, y, filterCoeff, sqrFiltLength);
          if (error<errMin || minInd==-1){
            errMin=error;
            minInd=k;
          }
        } // if (weights(k)<=diff){
      } // for (k=0; k<sqrFiltLength; k++){
      filterCoeffQuant[minInd]--;
    }
    else{
      diff=targetCoeffSumInt-quantCoeffSum;
      errMin=0; minInd=-1;
      for (k=0; k<sqrFiltLength; k++){
        if (weights[k]<=diff){
          for (i=0; i<sqrFiltLength; i++){
            filterCoeffQuantMod[i]=filterCoeffQuant[i];
          }
          filterCoeffQuantMod[k]++;
          for (i=0; i<sqrFiltLength; i++){
            filterCoeff[i]=(double)filterCoeffQuantMod[i]/(double)factor;
          }
          error=calculateErrorCoeffProvided(E, y, filterCoeff, sqrFiltLength);
          if (error<errMin || minInd==-1){
            errMin=error;
            minInd=k;
          }
        } // if (weights(k)<=diff){
      } // for (k=0; k<sqrFiltLength; k++){
      filterCoeffQuant[minInd]++;
    }

    quantCoeffSum=0;
    for (i=0; i<sqrFiltLength; i++){
      quantCoeffSum+=weights[i]*filterCoeffQuant[i];
    }
  }

   for (i=0; i<sqrFiltLength; i++){
    filterCoeff[i]=(double)filterCoeffQuant[i]/(double)factor;
  }

   error=calculateErrorCoeffProvided(E, y, filterCoeff, sqrFiltLength);
   return(error);

}

//
//
//
double findFilterCoeff(double ***EGlobalSeq, double **yGlobalSeq, double *pixAccGlobalSeq, double **filterCoeffSeq, int **filterCoeffQuantSeq, 
                       int intervalBest[NO_VAR_BINS][2], int varIndTab[NO_VAR_BINS], int sqrFiltLength, int filters_per_fr, int *weights, 
                       int bit_depth, double errorTabForce0Coeff[NO_VAR_BINS][2])
{
  static double **E_temp, *y_temp, *filterCoeff, pixAcc_temp;
  double error;
  static int *filterCoeffQuant;
  int k, filtNo;

  static int init = 0;

  if(init == 0)
  {
    initMatrix_double(&E_temp, MAX_SQR_FILT_LENGTH, MAX_SQR_FILT_LENGTH);
    y_temp = (double *) calloc(MAX_SQR_FILT_LENGTH, sizeof(double));
    filterCoeff = (double *) calloc(MAX_SQR_FILT_LENGTH, sizeof(double));
    filterCoeffQuant = (int *) calloc(MAX_SQR_FILT_LENGTH, sizeof(int));
    init = 1;
  }

  error = 0;
  for(filtNo = 0; filtNo < filters_per_fr; filtNo++)
  {
    add_A(E_temp, EGlobalSeq, intervalBest[filtNo][0], intervalBest[filtNo][1], sqrFiltLength);
    add_b(y_temp, yGlobalSeq, intervalBest[filtNo][0], intervalBest[filtNo][1], sqrFiltLength);

    pixAcc_temp = 0;    
    for(k = intervalBest[filtNo][0]; k <= intervalBest[filtNo][1]; k++)
      pixAcc_temp += pixAccGlobalSeq[k];

    // Find coeffcients
    errorTabForce0Coeff[filtNo][1] = pixAcc_temp + QuantizeIntegerFilterPP(filterCoeff, filterCoeffQuant, E_temp, y_temp, sqrFiltLength, weights, bit_depth);
    errorTabForce0Coeff[filtNo][0] = pixAcc_temp;
    error += errorTabForce0Coeff[filtNo][1];

    for(k = 0; k < sqrFiltLength; k++)
    {
      filterCoeffSeq[filtNo][k] = filterCoeff[k];
      filterCoeffQuantSeq[filtNo][k] = filterCoeffQuant[k];
    }
  }

  for(filtNo = 0; filtNo < filters_per_fr; filtNo++)
    for(k = intervalBest[filtNo][0]; k <= intervalBest[filtNo][1]; k++)
      varIndTab[k] = filtNo;

  return(error);
}
